/**
 * FXE - Formula eXpression Evaluator
 * 
 * ABOUT
 * 
 * Parses simple numeric formulas.
 * - Respects BEDMAS order of operations
 * - Functions; IF(cond, true [, false]), MIN(...), MAX(...), OR(...), AND(...)
 * - Division by zero produces a zero
 * 
 * USAGE
 * 
 * // browser
 * <script src="fxe.js"></script>
 * fxe.FXEngine.exec('1 + 2'); // returns 3
 * 
 * // node.js
 * var fxe = require('fxe');
 * assert(fxe.FXEngine.exec('1 + 2') === 3);
 * 
 * LICENSE
 * 
 * FXE Copyright (c) 2013, The FXE Author.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice and
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice
 *   and this list of conditions in the documentation and/or other materials
 *   provided with the distribution.
 * 
 */

;(function() {
	
	function FXToken(type, data) {
		this.type = parseInt(type, 10);
		this.data = data;
		this.children = [];
	}

	// Constants are static
	
	FXToken.TYPE_ERROR = 0;
	FXToken.TYPE_FNCALL = 1;
	FXToken.TYPE_ARGSEP = 2;
	FXToken.TYPE_OBRACKET = 3;
	FXToken.TYPE_CBRACKET = 4;
	FXToken.TYPE_NUMERIC = 5;
	FXToken.TYPE_PLUS = 6;
	FXToken.TYPE_MINUS = 7;
	FXToken.TYPE_GT = 8;
	FXToken.TYPE_LT = 9;
	FXToken.TYPE_EQ = 10;
	FXToken.TYPE_MUL = 11;
	FXToken.TYPE_DIV = 12;
	
	// Methods
	
	FXToken.prototype.makeError = function(sz) {
		this.type = FXToken.TYPE_ERROR;
		this.data = sz;
	};
	
	FXToken.prototype.makeNumeric = function(value) {
		this.type = FXToken.TYPE_NUMERIC;
		this.data = parseFloat(value, 10);
	};
	
	FXToken.prototype.getNumeric = function() {
		return (this.type === FXToken.TYPE_NUMERIC) ? this.data : null;
	};
	
	FXToken.prototype.getError = function() {
		return (this.type === FXToken.TYPE_ERROR) ? this.data : null;
	};
	
	/**
	 * Add a complete set of child parameters to this token. Would be called
	 *   exactly once for a bracket node, or zero+ times for a fncall node.
	 * 
	 * @param {FXToken[]} nest
	 * @returns {undefined}
	 */
	FXToken.prototype.absorb = function(nest) {
		var tmp = [];
		for (var i = 0, e = nest.length; i !== e; ++i) {
			if (nest[i].type === FXToken.TYPE_ARGSEP) {
				this.children.push( tmp.slice(0) ); // copy
				tmp = [];
			} else {
				tmp.push( nest[i] );
			}
		}
		this.children.push( tmp );
	};
	
	/**
	 * If this token has children (i.e. it's a bracket or function call), apply
	 *   all operations to reduce them into a single element.
	 * 
	 * @returns {undefined}
	 */
	FXToken.prototype.reduce = function() {
		if ( this.children.length === 0 ) return; // nothing to do
		var retlists = [];
		
		for (var cidx in this.children) {
			var clist = this.children[cidx]; // assignment by reference
			for (var i = 0, e = clist.length; i !== e; ++i) {
				this.children[cidx][i].reduce();
			}
			
			// Linear evaluate children
			if (clist.length) {
				var leval = FXToken.lin_eval(clist);
				if (leval.getError() !== null) {
					this.makeError( leval.getError() );
					return;
				}
				retlists.push( leval );
			}
		}
		
		if (this.type === FXToken.TYPE_FNCALL) {
			switch( this.data.toLowerCase() ) {
				
				case 'if': {
					if (retlists.length === 2) {
						if (retlists[0].getNumeric() === 1) {
							this.makeNumeric( retlists[1].getNumeric() );
						} else {
							this.makeNumeric( 0 ); // false
						}
						
					} else if (retlists.length === 3) {
						if (retlists[0].getNumeric() === 1) {
							this.makeNumeric( retlists[1].getNumeric() );
						} else {
							this.makeNumeric( retlists[2].getNumeric() );
						}
					
					} else {
						this.makeError('Unexpected number of arguments to if()');
					}
				} break;
				
				case 'max': {
					var max = null;
					for (var ridx = 0; ridx < retlists.length; ridx++) {
						if (max === null || retlists[ridx].getNumeric() > max) {
							max = retlists[ridx].getNumeric();
						}
					}
					this.makeNumeric( max === null ? 0 : max );
				} break;
				
				case 'min': {
					var min = null;
					for (var ridx = 0; ridx < retlists.length; ridx++) {
						if (min === null || retlists[ridx].getNumeric() < min) {
							min = retlists[ridx].getNumeric();
						}
					}
					this.makeNumeric( min === null ? 0 : min );
				} break;
				
				case 'or': {
					if (retlists.length === 0) {
						this.makeError('Unexpected number of arguments to or()');
					} else {
						this.makeNumeric(0);
						for (var ridx = 0; ridx < retlists.length; ridx++) {
							if (retlists[ridx].getNumeric() !== 0) this.makeNumeric(1);
						}
					}
				} break;
				
				case 'and': {
					if (retlists.length === 0) {
						this.makeError('Unexpected number of arguments to and()');
					} else {
						this.makeNumeric(1);
						for (var ridx = 0; ridx < retlists.length; ridx++) {
							if (retlists[ridx].getNumeric() === 0) this.makeNumeric(0);
						}
					}
				} break;
				
				default: {
					this.makeError("Unknown function \"" + this.data + "\"");
				} break;
			}
			this.children = [];
			return;
			
		} else if (retlists.length === 1) {

			// found result
			this.type = retlists[0].type;
			this.data = retlists[0].data;
			this.children = []; // Cleared
			
			return;
		}
		
		this.makeError("Unexpected number of nested elements");
		return;
	
	};
	
	// Static methods (attach to FXToken, not prototype)
	
	/**
	 * Applies numeric operators in correct precedence to a flattened list of
	 *   tokens (i.e. containing no nested braces or function calls)
	 * 
	 * @param {FXToken[]} tokens
	 * @returns {FXToken}
	 */
	FXToken.lin_eval = function(tokens) {		
		if (tokens.length % 2 === 0) {
			return new FXToken(FXToken.TYPE_ERROR, "Unexpected number of elements");
		}
		
		// Check errors and that all types are as expected
		var i;
		for (i = 0; i < tokens.length; i ++) {
			if (tokens[i].type === FXToken.TYPE_ERROR) {
				return tokens[i]; // copy error
			}
		}
		
		for (i = 0; i < tokens.length; i += 2) {
			if (tokens[i].type !== FXToken.TYPE_NUMERIC) {
				return new FXToken(FXToken.TYPE_ERROR, "Unexpected element (expected number)");
			}
		}
		
		for (i = 1; i < tokens.length; i += 2) {
			if (tokens[i].type === FXToken.TYPE_NUMERIC) {
				return new FXToken(FXToken.TYPE_ERROR, "Unexpected number");
			}
		}
		
		// Proceed to fold...
		for (i = 1; i < tokens.length;) {
			if (tokens[i].type === FXToken.TYPE_MUL) {
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data * tokens[i+1].data));
				
			} else if (tokens[i].type === FXToken.TYPE_DIV) {
				var ret = (tokens[i+1].data === 0) ? 0 : tokens[i-1].data / tokens[i+1].data;
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, ret));
				
			} else i += 2;
		}
		
		for (i = 1; i < tokens.length;) {
			if (tokens[i].type === FXToken.TYPE_PLUS) {
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data + tokens[i+1].data));
				
			} else if (tokens[i].type === FXToken.TYPE_MINUS) {
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data - tokens[i+1].data));
				
			} else i += 2;
		}
		
		for (i = 1; i < tokens.length;) {
			if (tokens[i].type === FXToken.TYPE_GT) {
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data > tokens[i+1].data ? 1 : 0));
				
			} else if (tokens[i].type === FXToken.TYPE_LT) {
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data < tokens[i+1].data ? 1 : 0));
				
			} else if (tokens[i].type === FXToken.TYPE_EQ) {
				tokens.splice(i-1, 3, new FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data === tokens[i+1].data ? 1 : 0));
				
			} else i += 2;
		}
		
		if (tokens.length === 1) {
			return tokens[0];
		} else {
			// Can't actually happen
			return new FXToken(FXToken.TYPE_ERROR, "Failed to completely evaluate operations");
		}
	};
	
	/**
	 * Lex a string into an array of FXToken objects.
	 * 
	 * @param {string} sz
	 * @returns {FXToken[]}
	 */
	FXToken.lex = function(sz) {
		var match = [];
		var ret = [];
		
		for(;;) {
			sz = sz.replace(/^[\s\r\n\t]+/, '');
			if (! sz.length) break;
			
			if (!! (match = sz.match(/^[a-zA-Z]+\s*\(/))) {
				ret.push(new FXToken(FXToken.TYPE_FNCALL, match[0].replace(/\s*\($/, '')));
				sz = sz.substr(match[0].length);
				
			} else if (sz[0] === ',') {
				ret.push(new FXToken(FXToken.TYPE_ARGSEP));
				sz = sz.substr(1);
				
			} else if (sz[0] === '(') {
				ret.push(new FXToken(FXToken.TYPE_OBRACKET));
				sz = sz.substr(1);
				
			} else if (sz[0] === ')') {
				ret.push(new FXToken(FXToken.TYPE_CBRACKET));
				sz = sz.substr(1);
				
			} else if (!! (match = sz.match(/^-?\d+\.?\d*/))) {
				ret.push(new FXToken(FXToken.TYPE_NUMERIC, parseFloat(match[0])));
				sz = sz.substr(match[0].length);
				
				
			} else if (sz[0] === '+') {
				ret.push(new FXToken(FXToken.TYPE_PLUS));
				sz = sz.substr(1);
				
			} else if (sz[0] === '-') {
				ret.push(new FXToken(FXToken.TYPE_MINUS));
				sz = sz.substr(1);
				
			} else if (sz[0] === '*') {
				ret.push(new FXToken(FXToken.TYPE_MUL));
				sz = sz.substr(1);
				
			} else if (sz[0] === '/') {
				ret.push(new FXToken(FXToken.TYPE_DIV));
				sz = sz.substr(1);
				
			} else if (sz[0] === '>') {
				ret.push(new FXToken(FXToken.TYPE_GT));
				sz = sz.substr(1);
				
			} else if (sz[0] === '<') {
				ret.push(new FXToken(FXToken.TYPE_LT));
				sz = sz.substr(1);
				
			} else if (sz[0] === '=') {
				ret.push(new FXToken(FXToken.TYPE_EQ));
				sz = sz.substr(1);
				
			} else {
				return [
					new FXToken(FXToken.TYPE_ERROR, 'Unexpected token in source near "'+sz.substr(0, 6)+'"')
				];
			}
		}
		
		return ret;
	};
	
	/**
	 * Partially convert a lexed list of tokens into a nested abstract syntax
	 *   tree.
	 * 
	 * @param {FXToken[]} tokens
	 * @param {int} pos
	 * @returns {FXToken[]}
	 */
	FXToken.parse = function(tokens, pos) {
		var ret = [], i = pos;
		for (var i = pos, e = tokens.length; i < e; ++i) {
			
			if (tokens[i].type === FXToken.TYPE_FNCALL || tokens[i].type === FXToken.TYPE_OBRACKET) {
				var subparse = FXToken.parse(tokens, i+1);
				tokens[i].absorb( subparse[0] );
				ret.push( tokens[i] );
				i += subparse[1] + 1;
				
			} else if (tokens[i].type === FXToken.TYPE_CBRACKET) {
				return [ret, i-pos];
				
			} else {
				ret.push( tokens[i] );
			}
		}
		
		// Reached end of parse
		return [ret, i-pos];
	};
	
	// Export FXEngine only
	
	var FXEngine = new Object;
	
	/**
	 * Evaluate a string of source code and return the numeric result. Returns
	 *   an error string on failure.
	 * 
	 * @param {String} str
	 * @returns {String|int}
	 */
	FXEngine.exec = function(str) {
		var lex = FXToken.lex("(" + str + ")");
		
		for (var i = 0; i < lex.length; i++) {
			var err = lex[i].getError();
			if (err !== null) return err;
		}
		
		var parsed = FXToken.parse( lex, 0 );
		
		if (lex.length - parsed[1] !== 0) {
			// Parsing the formula terminated without reaching all tokens
			return "Brace mismatch";
		}
		
		if (parsed[0].length !== 1) {
			// Multiple top-level results
			return "Brace mismatch";
		}
		
		var head = parsed[0][0];
		head.reduce();
		
		var ret = head.getNumeric();
		return (ret === null) ? head.getError() : ret;
	};
	
	if (typeof module !== 'undefined' && typeof module.exports !== 'undefined') {
		module.exports = FXEngine;
	} else {
		window.FXEngine = FXEngine;
	}
	
})();
