<?php

/**
 * FXE - Formula eXpression Evaluator
 * 
 * ABOUT
 * 
 * Parses simple numeric formulas.
 * - Respects BEDMAS order of operations
 * - Functions; IF(cond, true [, false]), MIN(...), MAX(...), OR(...), AND(...)
 * - Division by zero produces a zero
 * 
 * USAGE
 * 
 * assert(FXEngine::exec('1 + 2') === 3);
 * 
 * LICENSE
 * 
 * FXE Copyright (c) 2013, The FXE Author.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice and
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice
 *   and this list of conditions in the documentation and/or other materials
 *   provided with the distribution.
 * 
 */

class FXToken {
	const TYPE_ERROR    = 0;
	const TYPE_FNCALL   = 1;
	const TYPE_ARGSEP   = 2;
	const TYPE_OBRACKET = 3;
	const TYPE_CBRACKET = 4;
	const TYPE_NUMERIC  = 5;
	const TYPE_PLUS     = 6;
	const TYPE_MINUS    = 7;
	const TYPE_GT       = 8;
	const TYPE_LT       = 9;
	const TYPE_EQ       = 10;
	const TYPE_MUL      = 11;
	const TYPE_DIV      = 12;
	
	private $type;
	private $data;	
	private $children = array();
	
	public function __construct($type, $data=null) {
		$this->type = intval($type);
		$this->data = $data;
	}
	
	private function makeError($sz) {
		$this->type = self::TYPE_ERROR;
		$this->data = $sz;
		$this->children = array();
	}
	
	private function makeNumeric($value=0) {
		$this->type = self::TYPE_NUMERIC;
		$this->data = $value;
		$this->children = array();
	}
	
	public function getNumeric() {
		return ($this->type == self::TYPE_NUMERIC) ? $this->data : false;
	}
	
	public function getError() {
		return ($this->type == self::TYPE_ERROR) ? $this->data : false;
	}
	
	/**
	 * Add a complete set of child parameters to this token. Would be called
	 *   exactly once for a bracket node, or zero-or-many times for a function-
	 *   -call node.
	 * 
	 * @param array $nest
	 */
	private function absorb(array $nest) {
		$tmp = array();
		for ($i = 0, $e = count($nest); $i != $e; ++$i) {
			if ($nest[$i]->type == self::TYPE_ARGSEP) {
				$this->children[] = $tmp;
				$tmp = array();
			} else {
				$tmp[] = $nest[$i];
			}
		}
		$this->children[] = $tmp;
	}
	
	/**
	 * If this token has children (i.e. it's a bracket or function call), apply
	 *   all operations to reduce them into a single element.
	 * 
	 */
	public function reduce() {
		
		if (count($this->children) == 0) return; // nothing to do
		
		$retlists = array();
		
		foreach($this->children as $cidx => $clist) {
			for ($i = 0, $e = count($clist); $i != $e; ++$i) {
				$this->children[$cidx][$i]->reduce();
			}
			
			// linear evaluate children
			if (count($clist)) {
				$leval = self::lin_eval($clist);
				if ($leval->getError() !== false) {
					$this->makeError($leval->getError());
					return;
				}
				$retlists[] = $leval;
			}
		}
		
		if ($this->type == self::TYPE_FNCALL) {
			// ... switch fnname...
			
			switch(strtolower($this->data)) {
				
				case 'if': {
					if (count($retlists) == 2) {
						if ($retlists[0]->getNumeric() === 1) {
							$this->makeNumeric( $retlists[1]->getNumeric() );
						} else {
							$this->makeNumeric(0); // false
						}
						
					} else if (count($retlists) == 3) {
						if ($retlists[0]->getNumeric() === 1) {
							$this->makeNumeric( $retlists[1]->getNumeric() );
						} else {
							$this->makeNumeric( $retlists[2]->getNumeric() );
						}
						
					} else {
						$this->makeError('Unexpected number of arguments to if()');
					}
				} break;
				
				case 'max': {
					$max = null;
					foreach($retlists as $retk) {
						if (is_null($max) || $retk->getNumeric() > $max) {
							$max = $retk->getNumeric();
						}
					}
					$this->makeNumeric(is_null($max) ? 0 : $max);
				} break;
				
				case 'min': {
					$min = null;
					foreach($retlists as $retk) {
						if (is_null($min) || $retk->getNumeric() < $min) {
							$min = $retk->getNumeric();
						}
					}
					$this->makeNumeric(is_null($min) ? 0 : $min);
				} break;
				
				case 'or': {
					if (! count($retlists)) {
						$this->makeError('Unexpected number of arguments to or()'); break;
					}
					$this->makeNumeric(0);
					foreach($retlists as $retk) {
						if ($retk->getNumeric() != 0) {
							$this->makeNumeric(1);
						}
					}
				} break;
				
				case 'and': {
					if (! count($retlists)) {
						$this->makeError('Unexpected number of arguments to and()'); break;
					} 
					$this->makeNumeric(1);
					foreach($retlists as $retk) {
						if ($retk->getNumeric() == 0) {
							$this->makeNumeric(0);
						}
					}
					
				} break;
				
				default: {
					$this->makeError('Unknown function "'.$this->data.'"');
					
				} break;
			
			}
			$this->children = array(); // 
			return;
			
		} else if (count($retlists) === 1) {
			
			// found result
			$this->type     = $retlists[0]->type;
			$this->data     = $retlists[0]->data;
			$this->children = array(); // Cleared
					
			return;
		}
		
		$this->makeError('Unexpected number of nested elements'); // count(retlists) == 0 or, somehow > 1
		return;
		
	}
	
	/**
	 * PHP doesn't support macros. Internal flow control for lin_eval()
	 * 
	 * @param array $tokens
	 * @param int $i
	 * @param type $val
	 */
	private static function inject(array &$tokens, &$i, $val) {
		array_splice($tokens, $i-1, 3, array(new FXToken(self::TYPE_NUMERIC, $val)) );		
		$i -= 2;
	}
	
	/**
	 * Applies numeric operators in correct precendence to a flattened list
	 *   of tokens (i.e. containing no braces or function calls)
	 * 
	 * @param \FXToken[] $tokens
	 * @return \FXToken |false
	 */
	private static function lin_eval($tokens) {		
		if (count($tokens) % 2 == 0) {
			return new FXToken(self::TYPE_ERROR, 'Unexpected number of elements');
		}
		
		// check errors..
		for($i = 0; $i < count($tokens); $i++) {
			if ($tokens[$i]->type == self::TYPE_ERROR) {
				return $tokens[$i]; // copy error
			}
		}
		
		// assert odd-numbered options are all numeric..
		for($i = 0; $i < count($tokens); $i += 2) {
			if ($tokens[$i]->type != self::TYPE_NUMERIC) {
				return new FXToken(self::TYPE_ERROR, 'Unexpected element (expected number)');
			}
		}
		
		// and even-numbered options are all non-numeric
		for($i = 1; $i < count($tokens); $i += 2) {
			if ($tokens[$i]->type == self::TYPE_NUMERIC) {
				return new FXToken(self::TYPE_ERROR, 'Unexpected number');
			}
		}
		
		// proceed to fold
		
		for($i = 1; $i < count($tokens); $i += 2) {
			
			if ($tokens[$i]->type == self::TYPE_MUL) {
				self::inject($tokens, $i, $tokens[$i-1]->data * $tokens[$i+1]->data);
								
			} else if ($tokens[$i]->type == self::TYPE_DIV) {
				$res = ($tokens[$i+1]->data == 0) ? 0 : $tokens[$i-1]->data / $tokens[$i+1]->data;
				self::inject($tokens, $i, $res);
				
			}			
		}
		
		for($i = 1; $i < count($tokens); $i += 2) {
			
			if ($tokens[$i]->type == self::TYPE_PLUS) {
				self::inject($tokens, $i, $tokens[$i-1]->data + $tokens[$i+1]->data);
								
			} else if ($tokens[$i]->type == self::TYPE_MINUS) {
				self::inject($tokens, $i, $tokens[$i-1]->data - $tokens[$i+1]->data);
				
			}
			
		}
				
		for($i = 1; $i < count($tokens); $i += 2) {		
			if ($tokens[$i]->type == self::TYPE_GT) {
				self::inject($tokens, $i, $tokens[$i-1]->data > $tokens[$i+1]->data ? 1 : 0);
								
			} else if ($tokens[$i]->type == self::TYPE_LT) {
				self::inject($tokens, $i, $tokens[$i-1]->data < $tokens[$i+1]->data ? 1 : 0);
				
			} else if ($tokens[$i]->type == self::TYPE_EQ) {
				self::inject($tokens, $i, $tokens[$i-1]->data == $tokens[$i+1]->data ? 1 : 0);
				
			}
			
		}
		
		if (count($tokens) == 1) {
			return $tokens[0];
		} else {
			return new FXToken(self::TYPE_ERROR, 'Failed to completely evaluate operations'); // Can't actually happen
		}
		
	}
	
	/**
	 * Lex a string into an array of FXToken objects.
	 * 
	 * @param string $sz
	 * @return \FXToken[] |false
	 */
	public static function lex($sz) {
		$matches = array();
		$ret = array();

		for(;;) {
			// trim whitepsace
			$sz = preg_replace('~^[\s\r\n\t]+~m', '', $sz);
			if (! strlen($sz)) break;

			if (preg_match('~^([a-zA-Z]+\s*\()~', $sz, $matches)) {
				$ret[] = new FXToken(self::TYPE_FNCALL, trim(substr($matches[1], 0, strlen($matches[1])-1)) );
				$sz = substr($sz, strlen($matches[1]));
				
			} else if ($sz[0] == ',') {
				$ret[] = new FXToken(self::TYPE_ARGSEP);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '(') {
				$ret[] = new FXToken(self::TYPE_OBRACKET);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == ')') {
				$ret[] = new FXToken(self::TYPE_CBRACKET);
				$sz = substr($sz, 1);
				
			} else if (preg_match('~^(-?\d+\.?\d*)~', $sz, $matches)) {
				$ret[] = new FXToken(self::TYPE_NUMERIC, 0 + $matches[1]);
				$sz = substr($sz, strlen($matches[1]));
				
			} else if ($sz[0] == '+') {
				$ret[] = new FXToken(self::TYPE_PLUS);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '-') {
				$ret[] = new FXToken(self::TYPE_MINUS);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '*') {
				$ret[] = new FXToken(self::TYPE_MUL);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '/') {
				$ret[] = new FXToken(self::TYPE_DIV);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '>') {
				$ret[] = new FXToken(self::TYPE_GT);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '<') {
				$ret[] = new FXToken(self::TYPE_LT);
				$sz = substr($sz, 1);
				
			} else if ($sz[0] == '=') {
				$ret[] = new FXToken(self::TYPE_EQ);
				$sz = substr($sz, 1);
				
			} else {
				return array(
					new FXToken(self::TYPE_ERROR, 'Unexpected token in source near "'.substr($sz, 0, 6).'"')
				);
			}
		}

		return $ret;
	}
	
	/**
	 * Partially convert a lexed list of tokens into a nested abstract syntax
	 *   tree.
	 * 
	 * @param \FXToken[] $tokens
	 * @return array
	 */
	public static function parse(array $tokens, $pos = 0) {
		$ret = array();
		for($i = $pos, $e = count($tokens); $i < $e; ++$i) {

			if (
				$tokens[$i]->type == self::TYPE_FNCALL ||
				$tokens[$i]->type == self::TYPE_OBRACKET
			) {
				list($nest, $offset) = self::parse($tokens, $i+1);
				
				$tokens[$i]->absorb($nest);
				$ret[] = $tokens[$i];
				$i += $offset + 1;

			} else if ($tokens[$i]->type == self::TYPE_CBRACKET) {
				return array($ret, $i - $pos);

			} else {
				$ret[] = $tokens[$i];
			}
		}
		
		// Reached end of parse
		return array($ret, $i - $pos);
	}
	
}

class FXEngine {
	private function __construct() {} 
	
	/**
	 * Evaluate a string of source code and return the numeric result. Returns
	 *   an error string on failure.
	 * 
	 * @param string $str
	 * @return numeric|string
	 */
	public static function exec($str) {
		$lex = FXToken::lex('('.$str.')');
		$parsed = FXToken::parse( $lex );
		
		foreach($lex as $tok) {
			$err = $tok->getError();
			if ($err !== false) return $err;
		}
		
		if (count($lex) - $parsed[1] !== 0) {
			// Parsing formula terminated without reaching end of token list
			return 'Brace mismatch';
		}
		if (count($parsed[0]) !== 1) {
			// multiple top-level results. This only occurs if any user-provided
			//  braces extend one below our wrapper braces
			return 'Brace mismatch';
		}

		$head = $parsed[0][0];
		$head->reduce();

		$ret = $head->getNumeric();
		return ($ret === false) ? $head->getError() : $ret;
	}
}
