"""
FXE - Formula eXpression Evaluator

ABOUT

Parses simple numeric formulas.
- Respects BEDMAS order of operations
- Functions; IF(cond, true [, false]), MIN(...), MAX(...), OR(...), AND(...)
- Division by zero produces a zero

USAGE

import fxe
assert( fxe.FXEngine.run("1 + 2") == 3 );

LICENSE

FXE Copyright (c) 2013, The FXE Author.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice and
   this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice
   and this list of conditions in the documentation and/or other materials
   provided with the distribution.

"""

import re

class FXToken():
    TYPE_ERROR    = 0
    TYPE_FNCALL   = 1
    TYPE_ARGSEP   = 2
    TYPE_OBRACKET = 3
    TYPE_CBRACKET = 4
    TYPE_NUMERIC  = 5
    TYPE_PLUS     = 6
    TYPE_MINUS    = 7
    TYPE_GT       = 8
    TYPE_LT       = 9
    TYPE_EQ       = 10
    TYPE_MUL      = 11
    TYPE_DIV      = 12
    
    def __init__(self, type=TYPE_ERROR, data=None):
        self.type = type
        self.data = data
        self.children = []
        
    def makeError(self, sz):
        self.type = FXToken.TYPE_ERROR;
        self.data = ""+sz
        self.children = []
        
    def makeNumeric(self, val=0):
        self.type = FXToken.TYPE_NUMERIC;
        self.data = val
        self.children = []
        
    def getNumeric(self):
        return self.data if self.type == FXToken.TYPE_NUMERIC else None
        
    def getError(self):
        return self.data if self.type == FXToken.TYPE_ERROR else None
        
    def absorb(self, nest=[]):
        """Add a complete set of child parameters to this token. Would be called
        exactly once for a bracket node, or zero-or-many times for a function-call
        node."""
        
        tmp = []
        for i in range(0, len(nest)):
            if nest[i].type == FXToken.TYPE_ARGSEP:
                self.children.append(tmp)
                tmp = []
            else:
                tmp.append( nest[i] )
                
        self.children.append(tmp)
        
    def reduce(self):
        """If this token has children (i.e. it's a bracket or function call), apply
        all operations to reduce them into a single element."""
        
        if len(self.children) == 0:
            return; # nothing to do
        
        retlists = []
        
        for cidx, clist in enumerate(self.children):
            for i in range(0, len(clist)):
                self.children[cidx][i].reduce()
                
            # Linear evaluate children
            if len(clist) > 0:
                leval = FXToken.lin_eval(clist)
                if leval.getError() is not None:
                    self.makeError(leval.getError())
                    return
                retlists.append(leval)
                
        if self.type == FXToken.TYPE_FNCALL:
            
            sdl = self.data.lower() # Python doesn't have switch()
            if sdl == "if":
                if len(retlists) == 2:
                    if retlists[0].getNumeric() == 1:
                        self.makeNumeric( retlists[1].getNumeric() )
                    else:
                        self.makeNumeric(0) # False
                        
                elif len(retlists) == 3:
                    if retlists[0].getNumeric() == 1:
                        self.makeNumeric( retlists[1].getNumeric() )
                    else:
                        self.makeNumeric( retlists[2].getNumeric() )
                
                else:
                    self.makeError("Unexpected number of arguments to if()")
                
            elif sdl == "max":
                max = None
                for retk in retlists:
                    if max is None or retk.getNumeric() > max:
                        max = retk.getNumeric()
                        
                self.makeNumeric( 0 if max is None else max )
                
            elif sdl == "min":
                min = None
                for retk in retlists:
                    if min is None or retk.getNumeric() < min:
                        min = retk.getNumeric()
                
                self.makeNumeric( 0 if min is None else min )
                
            elif sdl == "or":
                if len(retlists) == 0:
                    self.makeError("Unexpected number of arguments to or()")
                else:
                    self.makeNumeric(0)
                    for retk in retlists:
                        if retk.getNumeric() != 0:
                            self.makeNumeric(1)
                
                
            elif sdl == "and":
                if len(retlists) == 0:
                    self.makeError("Unexpected number of arguments to and()")
                else:
                    self.makeNumeric(1)
                    for retk in retlists:
                        if retk.getNumeric() == 0:
                            self.makeNumeric(0)
                
            else:
                self.makeError("Unknown function \""+self.data+"\"")
                
            self.children = []
            return
                
        elif len(retlists) == 1:
            
            # Found result
            self.type = retlists[0].type
            self.data = retlists[0].data
            self.children = [] # Cleared
            
            return
        
        self.makeError("Unexpected number of nested elements") # count(retlists)==0 or somehow > 1
        return
    
    @staticmethod
    def lin_eval(tokens):
        if len(tokens) % 2 == 0:
            return FXToken(FXToken.TYPE_ERROR, "Unexpected number of elements")
        
        for i in range(0, len(tokens)):
            if tokens[i].type == FXToken.TYPE_ERROR:
                return tokens[i] # Copy error
            
        for i in range(0, len(tokens), 2):
            if tokens[i].type != FXToken.TYPE_NUMERIC:
                return FXToken(FXToken.TYPE_ERROR, "Unexpected element (expected number)")
            
        for i in range(1, len(tokens), 2):
            if tokens[i].type == FXToken.TYPE_NUMERIC:
                return FXToken(FXToken.TYPE_ERROR, "Unexpected number")
            
        # Proceed to fold
        
        i = 1
        while i < len(tokens):            
            if tokens[i].type == FXToken.TYPE_MUL:
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data * tokens[i+1].data),
                
            elif tokens[i].type == FXToken.TYPE_DIV:
                res = 0 if tokens[i+1].data == 0 else tokens[i-1].data / tokens[i+1].data
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, res),
                
            else:
                i += 2
                
        i = 1
        while i < len(tokens):
            if tokens[i].type == FXToken.TYPE_PLUS:
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data + tokens[i+1].data),
                
            elif tokens[i].type == FXToken.TYPE_MINUS:
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, tokens[i-1].data - tokens[i+1].data),
            
            else:
                i += 2
                
        i = 1
        while i < len(tokens):
            if tokens[i].type == FXToken.TYPE_GT:
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, 1 if tokens[i-1].data > tokens[i+1].data else 0),
            
            elif tokens[i].type == FXToken.TYPE_LT:
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, 1 if tokens[i-1].data < tokens[i+1].data else 0),
                
            elif tokens[i].type == FXToken.TYPE_EQ:
                tokens[i-1:i+2] = FXToken(FXToken.TYPE_NUMERIC, 1 if tokens[i-1].data == tokens[i+1].data else 0),
                
            else:
                i += 2
                
        if len(tokens) == 1:
            return tokens[0]
        else:
            return FXToken(FXToken.TYPE_ERROR, "Failed to completely evaluate operations") # Can't actually happen
            
    @staticmethod 
    def lex(sz):
        """Lex a string into an array of FXToken objects."""
        
        match = [None] # Wrap string in list to allow mutation by nested function...
        ret = []
        
        def rxmatch(rx, str):
            """Apply a regular expression match, put result in outer variable $match"""
            rm = re.compile(rx).match(str)
            if rm is None:
                match[0] = None
                return False
            else:
                match[0] = rm.group()
                return True
        
        while True:
            
            sz = re.compile("^[\\s\\r\\n\\t]+").sub("", sz)
            if len(sz) == 0:
                break
            
            if rxmatch("^[a-zA-Z]+\\s*\\(", sz):
                ret.append( FXToken(FXToken.TYPE_FNCALL, match[0][0:-1].strip() ) )
                sz = sz[len(match[0]):] 
            
            elif sz[0] == ",":
                ret.append( FXToken(FXToken.TYPE_ARGSEP) )
                sz = sz[1:]
                
            elif sz[0] == "(":
                ret.append( FXToken(FXToken.TYPE_OBRACKET) )
                sz = sz[1:]
                
            elif sz[0] == ")":
                ret.append( FXToken(FXToken.TYPE_CBRACKET) )
                sz = sz[1:]
                
            elif rxmatch("^-?\\d+\\.?\\d*", sz):
                ret.append( FXToken(FXToken.TYPE_NUMERIC, float(match[0]) ) )
                sz = sz[len(match[0]):]
                
            elif sz[0] == "+":
                ret.append( FXToken(FXToken.TYPE_PLUS) )
                sz = sz[1:]
                
            elif sz[0] == "-":
                ret.append( FXToken(FXToken.TYPE_MINUS) )
                sz = sz[1:]
                
            elif sz[0] == "*":
                ret.append( FXToken(FXToken.TYPE_MUL) )
                sz = sz[1:]
                
            elif sz[0] == "/":
                ret.append( FXToken(FXToken.TYPE_DIV) )
                sz = sz[1:]
                
            elif sz[0] == ">":
                ret.append( FXToken(FXToken.TYPE_GT) )
                sz = sz[1:]
                
            elif sz[0] == "<":
                ret.append( FXToken(FXToken.TYPE_LT) )
                sz = sz[1:]
                
            elif sz[0] == "=":
                ret.append( FXToken(FXToken.TYPE_EQ) )
                sz = sz[1:]
                
            else:
                return [
                    FXToken(FXToken.TYPE_ERROR, "Unexpected token in source near \"" + sz[:6] + "\"")
                ]
            
        return ret
           
    @staticmethod 
    def parse(tokens, pos=0):
        """Partially convert a lexed list of tokens into a nested abstract syntax tree."""
        
        ret = []
        i = pos
        while i < len(tokens):
            
            if tokens[i].type == FXToken.TYPE_FNCALL or tokens[i].type == FXToken.TYPE_OBRACKET:
                nest, offset = FXToken.parse(tokens, i+1) 
                
                tokens[i].absorb(nest)
                ret.append( tokens[i] )
                i += offset + 1
                
            elif tokens[i].type == FXToken.TYPE_CBRACKET:
                return [ret, i-pos]
            
            else:
                ret.append( tokens[i] )
                
            i += 1
               
        # Reached end of parse 
        return [ret, i-pos]
                
            
class FXEngine:
    
    @staticmethod
    def run(str):
        """Evaluate a string of source code and return the numeric result. Returns an error
        string on failure."""
        lex = FXToken.lex("(" + str + ")")
        
        for tok in lex:
            if tok.type == FXToken.TYPE_ERROR:
                return tok.getError()
        
        parsed = FXToken.parse(lex)
        
        if len(lex) - parsed[1] != 0:
            # Parsing formula terminated without reaching end of token list
            return "Brace mismatch"
        
        if len(parsed[0]) != 1:
            # Multiple top-level results - the user-provided braces extend only one below
            #  our wrapper braces
            return "Brace mismatch"
        
        head = parsed[0][0]
        head.reduce()
        
        ret = head.getNumeric()
        return head.getError() if ret == None else ret

            